# :coding: utf-8

import pytest
import os
import re
import unicodedata

from sphinx.cmdline import main as sphinx_main
from sphinx.util.osutil import cd


def _sanitise_value(value):
    """Return *value* suitable for comparison using python 2 and python 3.
    """
    value = value.decode("UTF-8")
    value = unicodedata.normalize("NFKD", value)
    value = value.encode("ascii", "ignore").decode("UTF-8")
    value = re.sub(
        r"[^\w*._\-\\/:% \"()\[\]{}\n=,]", "", value
    )
    return value


@pytest.fixture()
def doc_folder_with_code(doc_folder_with_class_members):
    """Return Doc folder with Javascript example source code.
    """
    js_source = os.path.join(doc_folder_with_class_members, "example")

    with open(os.path.join(js_source, "index.js"), "w") as f:
        f.write(
            "/**\n"
            " * A cool application.\n"
            " */\n"
            "\n"
            "import {\n"
            "    VARIABLE_OBJECT as ALIASED_VARIABLE_OBJECT\n"
            "} from './test_attribute';\n"
            "\n"
            "\n"
            "const undocumentedFunction = (arg) => {\n"
            "    console.log(arg)\n"
            "};\n"
            "\n"
            "\n"
            "/** A private function */\n"
            "function _privateFunction(arg) {\n"
            "    console.log(arg)\n"
            "}\n"
            "\n"
            "export ALIASED_VARIABLE_OBJECT;\n"
            "export * from './test_class';\n"
        )

    with open(os.path.join(js_source, "test_attribute.js"), "w") as f:
        f.write(
            "/**\n"
            " * A variable\n"
            " *\n"
            " * .. note::\n"
            " *\n"
            " *     A note.\n"
            " */\n"
            "export const VARIABLE_OBJECT = {\n"
            "    key1: 'value1',\n"
            "    key2: 'value2',\n"
            "    key3: 'value3',\n"
            "};\n"
            "\n"
            "/** Another variable. */\n"
            "let VARIABLE_INT = 42;\n"
        )

    with open(os.path.join(js_source, "test_class.js"), "w") as f:
        f.write(
            "/** A file with a great class. */\n"
            "\n"
            "\n"
            "import {Element as AliasedElement} from 'wherever';\n"
            "\n"
            "/**\n"
            " * Inherited class\n"
            " */\n"
            "class AwesomeClass {\n"
            "\n"
            "    /**\n"
            "     * Constructor.\n"
            "     */\n"
            "    constructor(name) {\n"
            "        super();\n"
            "        this.name = name;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Get name.\n"
            "     *\n"
            "     * .. warning::\n"
            "     *\n"
            "     *     The name is awesome\n"
            "     */\n"
            "    get name() {\n"
            "        return this.name;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Set name.\n"
            "     *\n"
            "     * .. warning::\n"
            "     *\n"
            "     *     Keep the name awesome\n"
            "     */\n"
            "    set name(value) {\n"
            "        this.name = value;\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * awesomeMethod.\n"
            "     */\n"
            "    awesomeMethod = () => {\n"
            "        console.log('Method has been called');\n"
            "    };\n"
            "\n"
            "    undocumentedMethod(arg1, arg2) {\n"
            "        console.log('An un-documented method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * Private Method.\n"
            "     */\n"
            "    _privateMethod(arg1) {\n"
            "        console.log('An private method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * staticMethod.\n"
            "     */\n"
            "    static staticMethod() {\n"
            "        console.log('Static method has been called');\n"
            "    }\n"
            "\n"
            "    /**\n"
            "     * attribute.\n"
            "     */\n"
            "    static attribute = 42;\n"
            "\n"
            "    /**\n"
            "     * another attribute.\n"
            "     */\n"
            "    classicAttribute = {\n"
            "        test: 'a test',\n"
            "    };\n"
            "}\n"
            "\n"
            "export AliasedElement;\n"
            "export default AwesomeClass;\n"
        )

    return doc_folder_with_class_members


def test_directive_automodule_error(doc_folder_with_code):
    """Do not generate doc for non existing module.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example.wrong"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "r"
    ) as f:
        assert f.read() == ""


def test_directive_automodule(doc_folder_with_code):
    """Generate documentation from modules.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "A file with a great class.\n"
        )


def test_directive_automodule_with_members(doc_folder_with_code):
    """Generate documentation from modules with members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_specific_members(doc_folder_with_code):
    """Generate documentation from modules with specific members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example.test_attribute\n"
            "    :members: VARIABLE_OBJECT\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
        )


def test_directive_automodule_with_undocumented_members(doc_folder_with_code):
    """Generate documentation from modules with undocumented members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :undoc-members:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :undoc-members:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :undoc-members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "example.undocumentedFunction(arg)\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_undocumented_members_default(
    doc_folder_with_code
):
    """Generate documentation from modules with undocumented members by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_module_options = ['undoc-members']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "example.undocumentedFunction(arg)\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_private_members(doc_folder_with_code):
    """Generate documentation from modules with private members.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :private-members:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :private-members:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :private-members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "example._privateFunction(arg)\n"
            "\n"
            "   A private function\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_private_members_default(
    doc_folder_with_code
):
    """Generate documentation from modules with private members by default.
    """
    conf_file = os.path.join(doc_folder_with_code, "conf.py")
    with open(conf_file, "a") as f:
        f.write("\njs_module_options = ['private-members']")

    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "example._privateFunction(arg)\n"
            "\n"
            "   A private function\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_module_alias(doc_folder_with_code):
    """Generate documentation from modules with module alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :module-alias: alias_module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const alias_module.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class alias_module.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const alias_module.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let alias_module.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class alias_module.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_module_path_alias(doc_folder_with_code):
    """Generate documentation from modules with module path alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :module-path-alias: alias/module\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :module-path-alias: alias/module\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :module-path-alias: alias/module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"alias/module\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"alias/module\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"alias/module\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"alias/module\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_partial_import_forced(doc_folder_with_code):
    """Generate documentation from modules with partial import forced.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :force-partial-import:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_data_value_skipped(doc_folder_with_code):
    """Generate documentation from modules with data value skipped.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :skip-data-value:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :skip-data-value:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :skip-data-value:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute = 42\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute = { test: a test, }\n"
            "\n"
            "      another attribute.\n"
        )


def test_directive_automodule_with_attribute_value_skipped(
    doc_folder_with_code
):
    """Generate documentation from modules with attribute value skipped.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:automodule:: example\n"
            "    :members:\n"
            "    :skip-attribute-value:\n"
            "\n"
            ".. js:automodule:: example.test_attribute\n"
            "    :members:\n"
            "    :skip-attribute-value:\n"
            "\n"
            ".. js:automodule:: example.test_class\n"
            "    :members:\n"
            "    :skip-attribute-value:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "A cool application.\n"
            "\n"
            "const example.ALIASED_VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_OBJECT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "class example.AwesomeClass(name)\n"
            "\n"
            "   \"import {AwesomeClass} from \"example\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute\n"
            "\n"
            "      another attribute.\n"
            "\n"
            "const example.test_attribute.VARIABLE_OBJECT = "
            "{ key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   \"import {VARIABLE_OBJECT} from \"example/test_attribute\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "let example.test_attribute.VARIABLE_INT = 42\n"
            "\n"
            "   Another variable.\n"
            "\n"
            "A file with a great class.\n"
            "\n"
            "class example.test_class.AwesomeClass(name)\n"
            "\n"
            "   \"import AwesomeClass from \"example/test_class\"\"\n"
            "\n"
            "   Inherited class\n"
            "\n"
            "   constructor(name)\n"
            "\n"
            "      Constructor.\n"
            "\n"
            "   get name()\n"
            "\n"
            "      Get name.\n"
            "\n"
            "      Warning: The name is awesome\n"
            "\n"
            "   set name(value)\n"
            "\n"
            "      Set name.\n"
            "\n"
            "      Warning: Keep the name awesome\n"
            "\n"
            "   awesomeMethod()\n"
            "\n"
            "      awesomeMethod.\n"
            "\n"
            "   static staticMethod()\n"
            "\n"
            "      staticMethod.\n"
            "\n"
            "   static attribute\n"
            "\n"
            "      attribute.\n"
            "\n"
            "   classicAttribute\n"
            "\n"
            "      another attribute.\n"
        )



# :coding: utf-8

import pytest
import os

import champollion.parser.helper


@pytest.mark.parametrize(
    ("content_lines", "line_number", "expected"),
    [
        (
            [
                "/**",
                " * An function example.",
                " *",
                " * Detailed description.",
                " *",
                " * .. note::",
                " *",
                " *     A note.",
                " */",
                "function sum(a, b) {",
                "    return a+b;",
                "}",
            ],
            10,
            (
                "An function example.\n"
                "\n"
                "Detailed description.\n"
                "\n"
                ".. note::\n"
                "\n"
                "    A note."
            )
        ),
        (
            [
                "/** A cool data. */",
                "const Data = null",
            ],
            2,
            (
                "A cool data."
            )
        ),
        (
            [
                "/*",
                " * Incorrect docstring",
                " */",
                "function doSomething() {",
                "    console.log('something');",
                "}",
            ],
            4,
            None
        ),
        (
            [
                "/*",
                "",
                " Incorrect docstring",
                "",
                "*/",
                "function doSomethingElse() {",
                "    console.log('something_else');",
                "}",
            ],
            6,
            None
        ),
        (
            [
                "// Incorrect docstring",
                "function doSomethingElse() {",
                "    console.log('something_else');",
                "}",
            ],
            2,
            None
        ),
        (
            [
                "",
                "function doSomethingElse() {",
                "    console.log('something_else');",
                "}",
            ],
            2,
            None
        ),
        (
            [
                "/** A cool data. */",
                "const Data = null",
            ],
            1,
            None
        )
    ],
    ids=[
        "valid element line number with multiline docstring",
        "valid element line number with one line docstring",
        "valid element line number with incorrect docstring 1",
        "valid element line number with incorrect docstring 2",
        "valid element line number with incorrect docstring 3",
        "valid element line number with no docstring",
        "invalid line_number",
    ]
)
def test_get_docstrings(content_lines, line_number, expected):
    """Return docstrings from a element's line number."""
    assert champollion.parser.helper.get_docstring(
        line_number, content_lines
    ) == expected


def test_filter_comments():
    """Remove all comments from content"""
    content = (
        "'use strict'; /* a beautiful comment */\n"
        "\n"
        "/*\n"
        "a long comment that can take a lot of places so\n"
        "we put it on several lines.\n"
        "*/\n"
        "\n"
        "// a data docstring\n"
        "const DATA = 1;\n"
        "\n"
        "/**\n"
        " * Function docstring\n"
        " */\n"
        "function sum(a, b) {\n"
        "    // Return the sum of a and b\n"
        "    return a+b;\n"
        "}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    )

    expected = (
        "'use strict'; \n"
        "\n"
        "\n"
        "\n"
        "\n"
        "\n"
        "\n"
        "\n"
        "const DATA = 1;\n"
        "\n"
        "\n"
        "\n"
        "\n"
        "function sum(a, b) {\n"
        "    \n"
        "    return a+b;\n"
        "}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    )
    assert champollion.parser.helper.filter_comments(content) == expected


def test_filter_comments_keep_content_size():
    """Remove all comments from content while keeping content size."""
    content = (
        "'use strict'; /* a beautiful comment */\n"
        "\n"
        "/*\n"
        "a long comment that can take a lot of places so\n"
        "we put it on several lines.\n"
        "*/\n"
        "\n"
        "// a data docstring\n"
        "const DATA = 1;\n"
        "\n"
        "/**\n"
        " * Function docstring\n"
        " */\n"
        "function sum(a, b) {\n"
        "    // Return the sum of a and b\n"
        "    return a+b;\n"
        "}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    )

    expected = (
        "'use strict'; {comment1}\n"
        "\n"
        "{comment2}\n"
        "\n"
        "\n"
        "\n"
        "\n"
        "{comment3}\n"
        "const DATA = 1;\n"
        "\n"
        "{comment4}\n"
        "\n"
        "\n"
        "function sum(a, b) {{\n"
        "    {comment5}\n"
        "    return a+b;\n"
        "}}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    ).format(
        comment1=" " * len("/* a beautiful comment */"),
        comment2=" " * len(
            "/*"
            "a long comment that can take a lot of places so"
            "we put it on several lines."
            "*/"
        ),
        comment3=" " * len("// a data docstring"),
        comment4=" " * len(
            "/**"
            " * Function docstring"
            " */"
        ),
        comment5=" " * len("// Return the sum of a and b")
    )

    assert champollion.parser.helper.filter_comments(
        content, keep_content_size=True
    ) == expected


def test_filter_comments_without_multiline_comments():
    """Remove all comments from content without multiline comments."""
    content = (
        "'use strict'; /* a beautiful comment */\n"
        "\n"
        "/*\n"
        "a long comment that can take a lot of places so\n"
        "we put it on several lines.\n"
        "*/\n"
        "\n"
        "// a data docstring\n"
        "const DATA = 1;\n"
        "\n"
        "/**\n"
        " * Function docstring\n"
        " */\n"
        "function sum(a, b) {\n"
        "    // Return the sum of a and b\n"
        "    return a+b;\n"
        "}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    )

    expected = (
        "'use strict'; /* a beautiful comment */\n"
        "\n"
        "/*\n"
        "a long comment that can take a lot of places so\n"
        "we put it on several lines.\n"
        "*/\n"
        "\n"
        "\n"
        "const DATA = 1;\n"
        "\n"
        "/**\n"
        " * Function docstring\n"
        " */\n"
        "function sum(a, b) {\n"
        "    \n"
        "    return a+b;\n"
        "}\n"
        "\n"
        "const url = 'http://somewhere.com';\n"
        "\n"
    )
    assert champollion.parser.helper.filter_comments(
        content, filter_multiline_comment=False
    ) == expected


@pytest.mark.parametrize(
    ("content", "expected_content", "expected_collapsed_content"),
    [
        (
            "const emptyObject = {};",
            "const emptyObject = {};",
            {}
        ),
        (
            "let test = {a: 1, b: 2, c: 3};",
            "let test = {};",
            {
                1: "{a: 1, b: 2, c: 3}"
            }
        ),
        (
            (
                "const element = {\n"
                "    key1: value1,\n"
                "    key2: value2,\n"
                "    key3: value3,\n"
                "};\n"
                "\n"
                "function sum(a, b) {\n"
                "    return a+b\n"
                "}\n"
                "\n"
            ),
            (
                "const element = {}\n"
                "\n"
                "\n"
                "\n"
                ";\n"
                "\n"
                "function sum(a, b) {}\n"
                "\n"
                "\n"
                "\n"
            ),
            {
                1: (
                    "{\n"
                    "    key1: value1,\n"
                    "    key2: value2,\n"
                    "    key3: value3,\n"
                    "}"
                ),
                7: "{\n"
                   "    return a+b\n"
                   "}"
            }
        ),
        (
            (
                "class AwesomeClass {\n"
                "    constructor() {\n"
                "        this.data = 1;\n"
                "    }\n"
                "\n"
                "    increase() {\n"
                "        this.data += 1;\n"
                "    }\n"
                "}\n"
            ),
            (
                "class AwesomeClass {}\n"
                "\n"
                "\n"
                "\n"
                "\n"
                "\n"
                "\n"
                "\n"
                "\n"
            ),
            {
                1: (
                    "{\n"
                    "    constructor() {\n"
                    "        this.data = 1;\n"
                    "    }\n"
                    "\n"
                    "    increase() {\n"
                    "        this.data += 1;\n"
                    "    }\n"
                    "}"
                ),
                2: (
                    "{\n"
                    "        this.data = 1;\n"
                    "    }"
                ),
                6: (
                    "{\n"
                    "        this.data += 1;\n"
                    "    }"
                )
            }
        )
    ],
    ids=[
        "empty object",
        "simple object",
        "objects and functions on multiple lines",
        "nested class"
    ]
)
def test_collapse_all(content, expected_content, expected_collapsed_content):
    """Collapse all objects, classes and functions."""
    assert champollion.parser.helper.collapse_all(content) == (
        expected_content, expected_collapsed_content
    )



# :coding: utf-8

from docutils import nodes
import re


def javascript_reference(
    name, rawtext, text, lineno, inliner, options=None, content=None
):
    """Return link to an Javascript reference for *text*.

    *name* should be the role name used in the document.
    *rawtext* should be the entire markup snippet, including the role whilst
    *text* is just the text marked with the role.

    *lineno* should be the line number where rawtext appears in the input.
    *inliner* should be the calling inliner instance.

     *options* should be a mapping of additional options for further
     customisation and *content* a list of additional context that may be
     supplied for customisation.

    Return tuple of list of nodes to insert and list of system messages.

    """
    app = inliner.document.settings.env.app

    match = re.match(r"(?P<alias>.*)<(?P<reference>.+)>", text)
    if match.group("alias") is None:
        ref_value = text.strip()
        title = ref_value.rsplit("/", 1)[-1]
    else:
        ref_value = match.group("reference").strip()
        title = match.group("alias").strip()

    reference = (
        "{reference_url}/{id}"
        .format(
            reference_url=app.config.reference_url,
            id=ref_value,
        )
    )

    node = nodes.reference(rawtext, title, refuri=reference)
    return [node], []


def setup(app):
    """Install the plugin in the Sphinx *app* context."""
    app.add_role_to_domain("js", "external", javascript_reference)
    app.add_config_value(
        "reference_url",
        "https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference",
        "env"
    )



# :coding: utf-8

from docutils.statemachine import StringList


def get_rst_class_elements(
    environment, module_name, module_path_name, whitelist_names=None,
    undocumented_members=False, private_members=False,
    force_partial_import=False, skip_attribute_value=False, rst_elements=None
):
    """Return :term:`reStructuredText` from class elements within
    *environment*.

    *module_name* is the module alias that should be added to each
    directive.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *whitelist_names* is an optional list of element names that
    should be displayed exclusively.

    *undocumented_members* indicate whether undocumented element should be
    displayed.

    *private_members* indicate whether elements starting with an underscore
    should be displayed.

    *force_partial_import* indicate whether the import statement should
    force the partial import display if necessary.

    *skip_attribute_value* indicate whether attribute value within the class
    should not be displayed.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    if rst_elements is None:
        rst_elements = {}

    for class_environment in environment["class"].values():
        name = class_environment["name"]
        description = class_environment["description"]
        if description is None and not undocumented_members:
            continue

        if name.startswith("_") and not private_members:
            continue

        if whitelist_names is None or name in whitelist_names:
            line_number = class_environment["line_number"]

            extra_options = []
            if force_partial_import:
                extra_options = [":force-partial-import:"]
            if skip_attribute_value:
                extra_options.append(":skip-attribute-value:")

            rst_element = rst_generate(
                directive="autoclass",
                element_id=class_environment["id"],
                alias=class_environment["name"],
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )
            rst_elements[line_number] = [rst_element]

    return rst_elements


def get_rst_attribute_elements(
    class_environment, whitelist_names=None, blacklist_ids=None,
    undocumented_members=False, private_members=False, skip_value=False,
    rst_elements=None,
):
    """Return :term:`reStructuredText` from class attribute elements within
    *class_environment*.

    *whitelist_names* is an optional list of element names that
    should be displayed exclusively.

    *blacklist_ids* is an optional list of element identifiers that
    should not be displayed.

    *undocumented_members* indicate whether undocumented element should be
    displayed.

    *private_members* indicate whether elements starting with an underscore
    should be displayed.

    *skip_value* indicate whether the value should not be displayed.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    if rst_elements is None:
        rst_elements = {}

    # As a method can also be a variable, use the method directive
    # by default when available
    if blacklist_ids is None:
        blacklist_ids = []

    for attr_environment in class_environment["attribute"].values():
        if attr_environment["id"] in blacklist_ids:
            continue

        name = attr_environment["name"]
        description = attr_environment["description"]
        if description is None and not undocumented_members:
            continue

        if name.startswith("_") and not private_members:
            continue

        extra_options = []
        if skip_value:
            extra_options.append(":skip-value:")

        if whitelist_names is None or name in whitelist_names:
            line_number = attr_environment["line_number"]
            rst_element = rst_generate(
                directive="autoattribute",
                element_id=attr_environment["id"],
                extra_options=extra_options
            )
            rst_elements[line_number] = [rst_element]

    return rst_elements


def get_rst_method_elements(
    class_environment, whitelist_names=None, skip_constructor=False,
    undocumented_members=False, private_members=False, rst_elements=None
):
    """Return :term:`reStructuredText` from class method elements within
    *class_environment*.

    *whitelist_names* is an optional list of element names that
    should be displayed exclusively.

    *skip_constructor* indicate whether the class constructor should be
    displayed.

    *undocumented_members* indicate whether undocumented element should be
    displayed.

    *private_members* indicate whether elements starting with an underscore
    should be displayed.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    if rst_elements is None:
        rst_elements = {}

    for method_environment in class_environment["method"].values():
        name = method_environment["name"]
        if name == "constructor" and skip_constructor:
            continue

        description = method_environment["description"]
        if description is None and not undocumented_members:
            continue

        if name.startswith("_") and not private_members:
            continue

        if whitelist_names is None or name in whitelist_names:
            line_number = method_environment["line_number"]
            rst_element = rst_generate(
                directive="automethod",
                element_id=method_environment["id"],
            )
            rst_elements[line_number] = [rst_element]

    return rst_elements


def get_rst_function_elements(
    environment, module_name, module_path_name, whitelist_names=None,
    undocumented_members=False, private_members=False,
    force_partial_import=False, rst_elements=None
):
    """Return :term:`reStructuredText` from function elements within
    *environment*.

    *module_name* is the module alias that should be added to each
    directive.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *whitelist_names* is an optional list of element names that
    should be displayed exclusively.

    *undocumented_members* indicate whether undocumented element should be
    displayed.

    *private_members* indicate whether elements starting with an underscore
    should be displayed.

    *force_partial_import* indicate whether the import statement should
    force the partial import display if necessary.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    if rst_elements is None:
        rst_elements = {}

    for function_environment in environment["function"].values():
        name = function_environment["name"]
        description = function_environment["description"]
        if description is None and not undocumented_members:
            continue

        if name.startswith("_") and not private_members:
            continue

        if whitelist_names is None or name in whitelist_names:
            line_number = function_environment["line_number"]

            extra_options = []
            if force_partial_import:
                extra_options = [":force-partial-import:"]

            rst_element = rst_generate(
                directive="autofunction",
                element_id=function_environment["id"],
                alias=function_environment["name"],
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )
            rst_elements[line_number] = [rst_element]

    return rst_elements


def get_rst_data_elements(
    environment, module_name, module_path_name, whitelist_names=None,
    blacklist_ids=None, undocumented_members=False, private_members=False,
    force_partial_import=False, skip_value=False, rst_elements=None,
):
    """Return :term:`reStructuredText` from data elements within
    *environment*.

    *module_name* is the module alias that should be added to each
    directive.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *whitelist_names* is an optional list of element names that
    should be displayed exclusively.

    *blacklist_ids* is an optional list of element identifiers that
    should not be displayed.

    *undocumented_members* indicate whether undocumented element should be
    displayed.

    *private_members* indicate whether elements starting with an underscore
    should be displayed.

    *force_partial_import* indicate whether the import statement should
    force the partial import display if necessary.

    *skip_value* indicate whether the value should not be displayed.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    if rst_elements is None:
        rst_elements = {}

    # As a function can also be a variable, use the function directive
    # by default when available
    if blacklist_ids is None:
        blacklist_ids = []

    for data_environment in environment["data"].values():
        if data_environment["id"] in blacklist_ids:
            continue

        name = data_environment["name"]
        description = data_environment["description"]
        if description is None and not undocumented_members:
            continue

        if name.startswith("_") and not private_members:
            continue

        if whitelist_names is None or name in whitelist_names:
            line_number = data_environment["line_number"]

            extra_options = []
            if force_partial_import:
                extra_options.append(":force-partial-import:")
            if skip_value:
                extra_options.append(":skip-value:")

            rst_element = rst_generate(
                directive="autodata",
                element_id=data_environment["id"],
                alias=data_environment["name"],
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )
            rst_elements[line_number] = [rst_element]

    return rst_elements


def get_rst_export_elements(
    file_environment, environment, module_name, module_path_name,
    skip_data_value=False, skip_attribute_value=False, rst_elements=None
):
    """Return :term:`reStructuredText` from exported elements within
    *file_environment*.

    *environment* is the full :term:`Javascript` environment processed
    in :mod:`~champollion.parser`.

    *module_name* is the module alias that should be added to each
    directive.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *skip_data_value* indicate whether data value should not be displayed.

    *skip_attribute_value* indicate whether attribute value should not be
    displayed.

    *rst_elements* can be an initial dictionary that will be updated and
    returned.

    """
    export_environment = file_environment["export"]
    import_environment = file_environment["import"]

    if rst_elements is None:
        rst_elements = {}

    for _exported_env_id, _exported_env in export_environment.items():
        from_module_id = _exported_env["module"]
        line_number = _exported_env["line_number"]

        if line_number not in rst_elements.keys():
            rst_elements[line_number] = []

        name = _exported_env["name"]
        alias = _exported_env["alias"]
        if alias is None:
            alias = name

        # Update module origin and name from import if necessary
        if (from_module_id is None and
                _exported_env_id in import_environment.keys()):
            name = import_environment[_exported_env_id]["name"]
            from_module_id = import_environment[_exported_env_id]["module"]

        # Ignore element if the origin module can not be found
        if from_module_id not in environment["module"].keys():
            continue

        from_module_environment = environment["module"][from_module_id]
        from_file_id = from_module_environment["file_id"]
        from_file_env = environment["file"][from_file_id]

        if name == "default":
            rst_element = get_rst_default_from_file_environment(
                from_file_env, alias, module_name, module_path_name,
                skip_data_value=skip_data_value,
                skip_attribute_value=skip_attribute_value,
            )
            if rst_element is None:
                continue

            rst_elements[line_number].append(rst_element)

        elif name == "*":
            extra_options = [
                ":force-partial-import:",
                ":members:",
                ":skip-description:"
            ]
            if skip_data_value:
                extra_options.append(":skip-data-value:")
            if skip_attribute_value:
                extra_options.append(":skip-attribute-value:")

            rst_element = rst_generate(
                directive="automodule",
                element_id=from_module_id,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )
            rst_elements[line_number].append(rst_element)

        else:
            rst_element = get_rst_name_from_file_environment(
                name, from_file_env, alias, module_name, module_path_name,
                skip_data_value=skip_data_value,
                skip_attribute_value=skip_attribute_value,
            )
            if rst_element is None:
                continue

            rst_elements[line_number].append(rst_element)

    return rst_elements


def get_rst_default_from_file_environment(
    file_environment, alias, module_name, module_path_name,
    skip_data_value=False, skip_attribute_value=False
):
    """Return :term:`reStructuredText` from default element in
    *file_environment*.

    *alias* is the name that should replace the element name.

    *module_name* is the module alias that should replace the element
    module name.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *skip_data_value* indicate whether data value should not be displayed.

    *skip_attribute_value* indicate whether attribute value should not be
    displayed.

    .. warning::

        Return None if no default is found in the file.

    """
    for class_env in file_environment["class"].values():
        if class_env["default"]:
            extra_options = [":force-partial-import:"]
            if skip_data_value:
                extra_options.append(":skip-data-value:")
            if skip_attribute_value:
                extra_options.append(":skip-attribute-value:")

            return rst_generate(
                directive="autoclass",
                element_id=class_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )

    for function_env in file_environment["function"].values():
        if function_env["default"]:
            return rst_generate(
                directive="autofunction",
                element_id=function_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=[":force-partial-import:"]
            )

    for data_env in file_environment["data"].values():
        if data_env["default"]:
            extra_options = [":force-partial-import:"]
            if skip_data_value:
                extra_options.append(":skip-value:")

            return rst_generate(
                directive="autodata",
                element_id=data_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )


def get_rst_name_from_file_environment(
    name, file_environment, alias, module_name, module_path_name,
    skip_data_value=False, skip_attribute_value=False
):
    """Return :term:`reStructuredText` element in *file_environment* from
    *name*.

    *alias* is the name that should replace the element name.

    *module_name* is the module name that should replace the element
    module name.

    *module_path_name* is the module path alias that should be added to each
    directive.

    *skip_data_value* indicate whether data value should not be displayed.

    *skip_attribute_value* indicate whether attribute value should not be
    displayed.

    .. warning::

        Return None if the element is not found in the file.

    """
    for class_env in file_environment["class"].values():
        if class_env["name"] == name and class_env["exported"]:
            extra_options = [":force-partial-import:"]
            if skip_data_value:
                extra_options.append(":skip-data-value:")
            if skip_attribute_value:
                extra_options.append(":skip-attribute-value:")

            return rst_generate(
                directive="autoclass",
                element_id=class_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )

    for function_env in file_environment["function"].values():
        if function_env["name"] == name and function_env["exported"]:
            return rst_generate(
                directive="autofunction",
                element_id=function_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=[":force-partial-import:"]
            )

    for data_env in file_environment["data"].values():
        if data_env["name"] == name and data_env["exported"]:
            extra_options = [":force-partial-import:"]
            if skip_data_value:
                extra_options.append(":skip-value:")

            return rst_generate(
                directive="autodata",
                element_id=data_env["id"],
                alias=alias,
                module_alias=module_name,
                module_path_alias=module_path_name,
                extra_options=extra_options
            )


def rst_generate(
    directive, element_id, alias=None, module_alias=None,
    module_path_alias=None, extra_options=None
):
    """Generate `StringList` from *directive* and *element_id*.

    *directive* is one of the directive added to the :term:`Javascript`
    domain by this sphinx extension.

    *element_id* is an element ID returned by the
    :mod:`~champollion.parser`.

    *alias* is the name that should replace the element name.

    *module_alias* is the module name that should replace the element
    module name.

    *module_path_alias* is the module path that should replace the element
    module path.

    *extra_options* can be a list of extra options to add to the directive.

    """
    if extra_options is None:
        extra_options = []

    element_rst = "\n.. js:{directive}:: {id}\n".format(
        directive=directive, id=element_id
    )

    if alias is not None:
        element_rst += "    :alias: {alias}\n".format(
            alias=alias
        )

    if module_alias is not None:
        element_rst += "    :module-alias: {module}\n".format(
            module=module_alias
        )

    if module_path_alias is not None:
        element_rst += "    :module-path-alias: {module}\n".format(
            module=module_path_alias
        )

    for option in extra_options:
        element_rst += "    {option}\n".format(
            option=option
        )

    element_rst += "\n"
    return StringList(element_rst.split("\n"))


def rst_string(expression=""):
    """Return `StringList` from *expression*.
    """
    return StringList(expression.split("\n"))



# :coding: utf-8

import pytest
import os

from sphinx.cmdline import main as sphinx_main
from sphinx.util.osutil import cd


def test_directive_error(doc_folder):
    """Do not generate doc for non existing data.
    """
    js_source = os.path.join(doc_folder, "example")
    with open(os.path.join(js_source, "index.js"), "w") as f:
        f.write(
            "/**\n"
            " * A variable\n"
            " *\n"
            " * .. note::\n"
            " *\n"
            " *     A note.\n"
            " */\n"
            "const VARIABLE = 42;\n"
        )

    index_file = os.path.join(doc_folder, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.UNEXISTING_VARIABLE"
        )

    with cd(doc_folder):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(os.path.join(doc_folder, "_build", "index.txt"), "r") as f:
        assert f.read() == ""



# :coding: utf-8

import pytest

import champollion.parser.js_data


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            (
                "/**\n"
                " * test dictionary data.\n"
                " * \n"
                " * Detailed description.\n"
                " */\n"
                "export const DATA = 42;\n"
            ),
            {
                "test.module.DATA": {
                    "id": "test.module.DATA",
                    "module_id": "test.module",
                    "exported": True,
                    "default": False,
                    "name": "DATA",
                    "type": "const",
                    "value": "42",
                    "line_number": 6,
                    "description": (
                        "test dictionary data.\n"
                        "\n"
                        "Detailed description."
                    )
                }
            }
        ),
        (
            (
                "export default var DATA = {\n"
                "   key1: value1,\n"
                "   key2: value2,\n"
                "   key3: value3,\n"
                "};\n"
            ),
            {
                "test.module.DATA": {
                    "id": "test.module.DATA",
                    "module_id": "test.module",
                    "exported": True,
                    "default": True,
                    "name": "DATA",
                    "type": "var",
                    "value": "{ key1: value1, key2: value2, key3: value3, }",
                    "line_number": 1,
                    "description": None
                }
            }
        ),
        (
            (
                "/** test list data */\n"
                "let DATA = [\n"
                "    1, 2, 3 ];\n"
            ),
            {
                "test.module.DATA": {
                    "id": "test.module.DATA",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "DATA",
                    "type": "let",
                    "value": "[ 1, 2, 3 ]",
                    "line_number": 2,
                    "description": "test list data"
                }
            }
        ),
        (
            (
                "let DATA = arg =>\n"
                "    console.log(arg);\n"
            ),
            {
                "test.module.DATA": {
                    "id": "test.module.DATA",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "DATA",
                    "type": "let",
                    "value": "arg =>console.log(arg)",
                    "line_number": 1,
                    "description": None
                },
            }
        ),
        (
            (
                "const SUM = (arg1, arg2) => {\n"
                "    return arg1 + arg2;\n"
                "};"
            ),
            {
                "test.module.SUM": {
                    "id": "test.module.SUM",
                    "module_id": "test.module",
                    "exported": False,
                    "default": False,
                    "name": "SUM",
                    "type": "const",
                    "value": "(arg1, arg2) => { return arg1 + arg2; }",
                    "line_number": 1,
                    "description": None
                },
            }
        )
    ],
    ids=[
        "valid data",
        "valid object data",
        "valid list data",
        "valid arrow-type function with single argument",
        "valid arrow-type function with severak arguments",
    ]
)
def test_get_data_environment(content, expected):
    """Return data environment from content."""
    assert champollion.parser.js_data.fetch_environment(
        content, "test.module"
    ) == expected


@pytest.mark.parametrize(
    ("content", "expected"),
    [
        (
            "const data_test = 42;",
            {
                "name": "data_test",
                "type": "const",
                "value": "42;",
                "export": None,
                "default": None,
                "start_regex": ""
            }
        ),
        (
            (
                "export let data_test2 = {\n"
                "    key: 'value',\n"
                "};"
            ),
            {
                "name": "data_test2",
                "type": "let",
                "value": (
                    "{\n"
                    "    key: 'value',\n"
                    "};"
                ),
                "export": "export ",
                "default": None,
                "start_regex": ""
            }
        ),
        (
            (
                "var dataTest3 = [\n"
                "    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,\n"
                "];"
            ),
            {
                "name": "dataTest3",
                "type": "var",
                "value": (
                    "[\n"
                    "    1, 2, 3, 4, 5, 6, 7, 8, 9, 10,\n"
                    "];"
                ),
                "export": None,
                "default": None,
                "start_regex": ""
            }
        ),
        (
            "'const attribute = 42'",
            None
        ),
        (
            "const data_test = 42",
            None
        )
    ],
    ids=[
        "valid data",
        "valid object data",
        "valid list data",
        "invalid data string",
        "invalid data with no semi-colons",
    ]
)
def test_data_pattern(content, expected):
    """Match an variable."""
    match = champollion.parser.js_data._DATA_PATTERN.search(
        content
    )
    if expected is None:
        assert match is None
    else:
        assert match.groupdict() == expected



# :coding: utf-8

import pytest
import os
import re
import unicodedata

from sphinx.cmdline import main as sphinx_main
from sphinx.util.osutil import cd


def _sanitise_value(value):
    """Return *value* suitable for comparison using python 2 and python 3.
    """
    value = value.decode("UTF-8")
    value = unicodedata.normalize("NFKD", value)
    value = value.encode("ascii", "ignore").decode("UTF-8")
    value = re.sub(
        r"[^\w._\-\\/:% \"()\[\]{}\n=,]", "", value
    )
    return value


@pytest.fixture()
def doc_folder_with_code(doc_folder):
    """Return Doc folder with Javascript example source code.
    """
    js_source = os.path.join(doc_folder, "example")

    with open(os.path.join(js_source, "index.js"), "w") as f:
        f.write(
            "/**\n"
            " * A variable\n"
            " *\n"
            " * .. note::\n"
            " *\n"
            " *     A note.\n"
            " */\n"
            "export default const VARIABLE_INT = 42;\n"
            "\n"
            "/**\n"
            " * Another variable\n"
            " *\n"
            " * A citation::\n"
            " *\n"
            " *     A citation\n"
            " */\n"
            "var VARIABLE_OBJECT = {\n"
            "    key1: 'value1',\n"
            "    key2: 'value2',\n"
            "    key3: 'value3',\n"
            "};\n"
            "\n"
            "export let VARIABLE_STRING = 'rosebud';\n"
        )

    return doc_folder


def test_directive_autodata(doc_folder_with_code):
    """Generate documentation from global data variables.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.VARIABLE_INT = 42\n"
            "\n"
            "   \"import VARIABLE_INT from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var example.VARIABLE_OBJECT = { "
            "key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let example.VARIABLE_STRING = rosebud\n"
            "\n"
            "   \"import {VARIABLE_STRING} from \"example\"\"\n"
        )


def test_directive_autodata_with_alias(doc_folder_with_code):
    """Generate documentation from global data variables with alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "    :alias: ALIASED_VARIABLE_INT\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "    :alias: ALIASED_VARIABLE_OBJECT\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
            "    :alias: ALIASED_VARIABLE_STRING\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.ALIASED_VARIABLE_INT = 42\n"
            "\n"
            "   \"import ALIASED_VARIABLE_INT from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var example.ALIASED_VARIABLE_OBJECT = { "
            "key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let example.ALIASED_VARIABLE_STRING = rosebud\n"
            "\n"
            "   \"import {ALIASED_VARIABLE_STRING} from \"example\"\"\n"
        )


def test_directive_autodata_with_module_alias(doc_folder_with_code):
    """Generate documentation from global data variables with module alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "    :module-alias: alias_module\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
            "    :module-alias: alias_module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const alias_module.VARIABLE_INT = 42\n"
            "\n"
            "   \"import VARIABLE_INT from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var alias_module.VARIABLE_OBJECT = { "
            "key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let alias_module.VARIABLE_STRING = rosebud\n"
            "\n"
            "   \"import {VARIABLE_STRING} from \"example\"\"\n"
        )


def test_directive_autodata_with_module_path_alias(doc_folder_with_code):
    """Generate documentation from global data variables with module path alias.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "    :module-path-alias: test/alias/module\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
            "    :module-path-alias: test/alias/module\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.VARIABLE_INT = 42\n"
            "\n"
            "   \"import VARIABLE_INT from \"test/alias/module\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var example.VARIABLE_OBJECT = { "
            "key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let example.VARIABLE_STRING = rosebud\n"
            "\n"
            "   \"import {VARIABLE_STRING} from \"test/alias/module\"\"\n"
        )


def test_directive_autodata_with_partial_import_forced(doc_folder_with_code):
    """Generate documentation from global data variables with partial import
    forced.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "    :force-partial-import:\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
            "    :force-partial-import:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.VARIABLE_INT = 42\n"
            "\n"
            "   \"import {VARIABLE_INT} from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var example.VARIABLE_OBJECT = { "
            "key1: value1, key2: value2, key3: value3, }\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let example.VARIABLE_STRING = rosebud\n"
            "\n"
            "   \"import {VARIABLE_STRING} from \"example\"\"\n"
        )


def test_directive_autodata_with_value_skipped(doc_folder_with_code):
    """Generate documentation from global data variables with value skipped.
    """
    index_file = os.path.join(doc_folder_with_code, "index.rst")
    with open(index_file, "w") as f:
        f.write(
            ".. js:autodata:: example.VARIABLE_INT\n"
            "    :skip-value:\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_OBJECT\n"
            "    :skip-value:\n"
            "\n"
            ".. js:autodata:: example.VARIABLE_STRING\n"
            "    :skip-value:\n"
        )

    with cd(doc_folder_with_code):
        sphinx_main(["-c", ".", "-b", "text", "-E", ".", "_build"])

    with open(
        os.path.join(doc_folder_with_code, "_build", "index.txt"), "rb"
    ) as f:
        content = _sanitise_value(f.read())

        assert content == (
            "const example.VARIABLE_INT\n"
            "\n"
            "   \"import VARIABLE_INT from \"example\"\"\n"
            "\n"
            "   A variable\n"
            "\n"
            "   Note: A note.\n"
            "\n"
            "var example.VARIABLE_OBJECT\n"
            "\n"
            "   Another variable\n"
            "\n"
            "   A citation:\n"
            "\n"
            "      A citation\n"
            "\n"
            "let example.VARIABLE_STRING\n"
            "\n"
            "   \"import {VARIABLE_STRING} from \"example\"\"\n"
        )



# :coding: utf-8

from sphinx import addnodes
import docutils.parsers.rst.directives

from .base import BaseDirective


class AutoDataDirective(BaseDirective):
    """Directive to render :term:`Javascript` data documentation.

    The unique argument should be the identifier of the data element.

    .. sourcecode:: rest

        .. js:autodata:: module.DATA

    The available options are:

    * alias:
        String element to replace the data name.

    * module-alias:
        String element to replace the module name.

    * module-path-alias:
        String element to replace the module path.

    * force-partial-import:
        Indicate whether the data import statement display should be indicated
        with partial import if the data element is exported.

    * skip-value:
        Indicate whether data value should be skipped.

    .. seealso::

        :ref:`directive/autodata`

    """
    #: Javascript data are not callable
    has_arguments = False

    #: Define the Object type
    objtype = "data"

    #: data options
    option_spec = {
        "alias": docutils.parsers.rst.directives.unchanged_required,
        "module-alias": docutils.parsers.rst.directives.unchanged_required,
        "module-path-alias": docutils.parsers.rst.directives.unchanged_required,
        "force-partial-import": lambda x: True,
        "skip-value": lambda x: True,
    }

    def handle_signature(self, signature, node):
        """Update the signature *node*."""
        env = self.state.document.settings.env.element_environment
        module_env = self.state.document.settings.env.module_environment

        name = self.options.get("alias", env["name"])
        module_id = env["module_id"]
        module_name = self.options.get(
            "module-alias", module_env[module_id]["name"]
        )
        value = env["value"]
        variable_type = env["type"]

        skip_value = self.options.get("skip-value", False)

        node["type"] = "data"
        node["id"] = env["id"]
        node["module"] = module_name
        node["fullname"] = name

        node += addnodes.desc_type(variable_type + " ", variable_type + " ")
        node += addnodes.desc_addname(module_name + ".", module_name + ".")
        node += addnodes.desc_name(name, name)
        if not skip_value:
            node += addnodes.desc_annotation(" = " + value, " = " + value)

        return name, module_name

    def before_content(self):
        """Update the content.

        Compute the description and import statement if available.

        """
        env = self.state.document.settings.env.element_environment
        module_env = self.state.document.settings.env.module_environment

        self.content = self.generate_import_statement(
            env, module_env, self.options.get("force-partial-import")
        )
        self.content += self.generate_description(env)



# :coding: utf-8

import pytest

import champollion.directive.rst_generator
from docutils.statemachine import StringList


@pytest.mark.parametrize(
    ("options", "expected"),
    [
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class4",
                            "    :alias: Class4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": ["Class1"],
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": True,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                8: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class2",
                            "    :alias: Class2",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class4",
                            "    :alias: Class4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": True,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                16: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test._Class3",
                            "    :alias: _Class3",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class4",
                            "    :alias: Class4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": True,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class4",
                            "    :alias: Class4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": {
                    42: [
                        StringList(
                            [
                                "",
                                ".. js:autoclass:: example.Test",
                                "    :alias: Test",
                                "    :module-alias: example",
                                "    :module-path-alias: example",
                                "",
                                ""
                            ]
                        )
                    ]
                }
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class1",
                            "    :alias: Class1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: test.Class4",
                            "    :alias: Class4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                42: [
                    StringList(
                        [
                            "",
                            ".. js:autoclass:: example.Test",
                            "    :alias: Test",
                            "    :module-alias: example",
                            "    :module-path-alias: example",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
    ],
    ids=[
        "no options",
        "with whitelist names",
        "with undocumented members",
        "with private members",
        "with force-partial-import option",
        "with initial RST elements"
    ]
)
def test_get_rst_class_elements(options, expected):
    """Generate reStructuredText dictionary from class environment.
    """
    environment = {
        "class": {
            "test.Class1": {
                "id": "test.Class1",
                "name": "Class1",
                "description": "A super class\n\n.. note:: A note.",
                "line_number": 2
            },
            "test.Class2": {
                "id": "test.Class2",
                "name": "Class2",
                "description": None,
                "line_number": 8
            },
            "test._Class3": {
                "id": "test._Class3",
                "name": "_Class3",
                "description": "A private class\n\n.. note:: A note.",
                "line_number": 16
            },
            "test.Class4": {
                "id": "test.Class4",
                "name": "Class4",
                "description": "An interesting class\n\n.. note:: A note.",
                "line_number": 25
            },
        }
    }

    assert champollion.directive.rst_generator.get_rst_class_elements(
        environment, "test.module", "test/module", **options
    ) == expected


@pytest.mark.parametrize(
    ("options", "expected"),
    [
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute1",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute4",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": ["attribute1"],
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute1",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": ["test.Class.attribute1"],
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute4",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": True,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute1",
                            "",
                            ""
                        ]
                    )
                ],
                8: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute2",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute4",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": True,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute1",
                            "",
                            ""
                        ]
                    )
                ],
                16: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class._attribute3",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute4",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": {
                    42: [
                        StringList(
                            [
                                "",
                                ".. js:autoattribute:: example.Test.attribute",
                                "",
                                ""
                            ]
                        )
                    ]
                }
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute1",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: test.Class.attribute4",
                            "",
                            ""
                        ]
                    )
                ],
                42: [
                    StringList(
                        [
                            "",
                            ".. js:autoattribute:: example.Test.attribute",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
    ],
    ids=[
        "no options",
        "with whitelist names",
        "with blacklist ids",
        "with undocumented members",
        "with private members",
        "with initial RST elements"
    ]
)
def test_get_rst_attribute_elements(options, expected):
    """Generate reStructuredText dictionary from class attribute environment.
    """
    class_environment = {
        "attribute": {
            "test.Class.attribute1": {
                "id": "test.Class.attribute1",
                "name": "attribute1",
                "description": "A super attribute\n\n.. note:: A note.",
                "line_number": 2
            },
            "test.Class.attribute2": {
                "id": "test.Class.attribute2",
                "name": "attribute2",
                "description": None,
                "line_number": 8
            },
            "test.Class._attribute3": {
                "id": "test.Class._attribute3",
                "name": "_attribute3",
                "description": "A private attribute\n\n.. note:: A note.",
                "line_number": 16
            },
            "test.Class.attribute4": {
                "id": "test.Class.attribute4",
                "name": "attribute4",
                "description": "An interesting attribute\n\n.. note:: A note.",
                "line_number": 25
            },
        }
    }

    assert champollion.directive.rst_generator.get_rst_attribute_elements(
        class_environment, **options
    ) == expected


@pytest.mark.parametrize(
    ("options", "expected"),
    [
        (
            {
                "whitelist_names": None,
                "skip_constructor": False,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.constructor",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": ["method1"],
                "skip_constructor": False,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "skip_constructor": True,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "skip_constructor": False,
                "undocumented_members": True,
                "private_members": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ],
                8: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method2",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.constructor",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "skip_constructor": False,
                "undocumented_members": False,
                "private_members": True,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ],
                16: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class._method3",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.constructor",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "skip_constructor": False,
                "undocumented_members": False,
                "private_members": False,
                "rst_elements": {
                    42: [
                        StringList(
                            [
                                "",
                                ".. js:automethod:: example.Test.method",
                                "",
                                ""
                            ]
                        )
                    ]
                }
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.method1",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: test.Class.constructor",
                            "",
                            ""
                        ]
                    )
                ],
                42: [
                    StringList(
                        [
                            "",
                            ".. js:automethod:: example.Test.method",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
    ],
    ids=[
        "no options",
        "with whitelist names",
        "with constructor skipped",
        "with undocumented members",
        "with private members",
        "with initial RST elements"
    ]
)
def test_get_rst_method_elements(options, expected):
    """Generate reStructuredText dictionary from class method environment.
    """
    class_environment = {
        "method": {
            "test.Class.method1": {
                "id": "test.Class.method1",
                "name": "method1",
                "description": "A super method\n\n.. note:: A note.",
                "line_number": 2
            },
            "test.Class.method2": {
                "id": "test.Class.method2",
                "name": "method2",
                "description": None,
                "line_number": 8
            },
            "test.Class._method3": {
                "id": "test.Class._method3",
                "name": "_method3",
                "description": "A private method\n\n.. note:: A note.",
                "line_number": 16
            },
            "test.Class.constructor": {
                "id": "test.Class.constructor",
                "name": "constructor",
                "description": "A constructor.",
                "line_number": 25
            },
        }
    }

    assert champollion.directive.rst_generator.get_rst_method_elements(
        class_environment, **options
    ) == expected


@pytest.mark.parametrize(
    ("options", "expected"),
    [
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function4",
                            "    :alias: function4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": ["function1"],
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": True,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                8: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function2",
                            "    :alias: function2",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function4",
                            "    :alias: function4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": True,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                16: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test._function3",
                            "    :alias: _function3",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function4",
                            "    :alias: function4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": True,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function4",
                            "    :alias: function4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": {
                    42: [
                        StringList(
                            [
                                "",
                                ".. js:autofunction:: example.test",
                                "    :alias: test",
                                "    :module-alias: example",
                                "    :module-path-alias: example",
                                "",
                                ""
                            ]
                        )
                    ]
                }
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function1",
                            "    :alias: function1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: test.function4",
                            "    :alias: function4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                42: [
                    StringList(
                        [
                            "",
                            ".. js:autofunction:: example.test",
                            "    :alias: test",
                            "    :module-alias: example",
                            "    :module-path-alias: example",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
    ],
    ids=[
        "no options",
        "with whitelist names",
        "with undocumented members",
        "with private members",
        "with force-partial-import option",
        "with initial RST elements"
    ]
)
def test_get_rst_function_elements(options, expected):
    """Generate reStructuredText dictionary from function environment.
    """
    environment = {
        "function": {
            "test.function1": {
                "id": "test.function1",
                "name": "function1",
                "description": "A super function\n\n.. note:: A note.",
                "line_number": 2
            },
            "test.function2": {
                "id": "test.function2",
                "name": "function2",
                "description": None,
                "line_number": 8
            },
            "test._function3": {
                "id": "test._function3",
                "name": "_function3",
                "description": "A private function\n\n.. note:: A note.",
                "line_number": 16
            },
            "test.function4": {
                "id": "test.function4",
                "name": "function4",
                "description": "An interesting function\n\n.. note:: A note.",
                "line_number": 25
            },
        }
    }

    assert champollion.directive.rst_generator.get_rst_function_elements(
        environment, "test.module", "test/module", **options
    ) == expected


@pytest.mark.parametrize(
    ("options", "expected"),
    [
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": ["DATA1"],
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": ["test.DATA1"],
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": True,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                8: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA2",
                            "    :alias: DATA2",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": True,
                "force_partial_import": False,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                16: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test._DATA3",
                            "    :alias: _DATA3",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": True,
                "rst_elements": None,
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "    :force-partial-import:",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
        (
            {
                "whitelist_names": None,
                "blacklist_ids": None,
                "undocumented_members": False,
                "private_members": False,
                "force_partial_import": False,
                "rst_elements": {
                    42: [
                        StringList(
                            [
                                "",
                                ".. js:autodata:: example.TEST",
                                "    :alias: TEST",
                                "    :module-alias: example",
                                "    :module-path-alias: example",
                                "",
                                ""
                            ]
                        )
                    ]
                }
            },
            {
                2: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA1",
                            "    :alias: DATA1",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                25: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: test.DATA4",
                            "    :alias: DATA4",
                            "    :module-alias: test.module",
                            "    :module-path-alias: test/module",
                            "",
                            ""
                        ]
                    )
                ],
                42: [
                    StringList(
                        [
                            "",
                            ".. js:autodata:: example.TEST",
                            "    :alias: TEST",
                            "    :module-alias: example",
                            "    :module-path-alias: example",
                            "",
                            ""
                        ]
                    )
                ]
            }
        ),
    ],
    ids=[
        "no options",
        "with whitelist names",
        "with blacklist ids",
        "with undocumented members",
        "with private members",
        "with force-partial-import option",
        "with initial RST elements"
    ]
)
def test_get_rst_data_elements(options, expected):
    """Generate reStructuredText dictionary from data environment.
    """
    environment = {
        "data": {
            "test.DATA1": {
                "id": "test.DATA1",
                "name": "DATA1",
                "description": "A super data\n\n.. note:: A note.",
                "line_number": 2
            },
            "test.DATA2": {
                "id": "test.DATA2",
                "name": "DATA2",
                "description": None,
                "line_number": 8
            },
            "test._DATA3": {
                "id": "test._DATA3",
                "name": "_DATA3",
                "description": "A private data\n\n.. note:: A note.",
                "line_number": 16
            },
            "test.DATA4": {
                "id": "test.DATA4",
                "name": "DATA4",
                "description": "An interesting data\n\n.. note:: A note.",
                "line_number": 25
            },
        }
    }

    assert champollion.directive.rst_generator.get_rst_data_elements(
        environment, "test.module", "test/module", **options
    ) == expected



# :coding: utf-8

import pytest

import champollion.parser.js_module


def test_get_module_environment_from_file():
    """Return module_id and environment from file id."""
    assert champollion.parser.js_module.fetch_environment(
        "test/module/example.js", []
    ) == {
        "id": "test.module.example",
        "name": "example",
        "path": "test/module/example",
        "file_id": "test/module/example.js"
    }


def test_get_module_environment_from_index_file():
    """Return module_id and environment from index file id."""
    assert champollion.parser.js_module.fetch_environment(
        "test/module/index.js", []
    ) == {
            "id": "test.module",
            "name": "module",
            "path": "test/module",
            "file_id": "test/module/index.js"
    }


def test_get_module_environment_from_file_with_adjacent_index():
    """Return module_id and environment from file id with adjacent index file.
    """
    assert champollion.parser.js_module.fetch_environment(
        "test/module/example.js", ["index.js"]
    ) == {
        "id": "test.module.example",
        "name": "module.example",
        "path": "test/module/example",
        "file_id": "test/module/example.js"
    }


@pytest.mark.parametrize(
    ("name", "hierarchy_folders", "module_names", "expected"),
    [
        (
            "example",
            ["module", "submodule", "test"],
            [],
            "example"
        ),
        (
            "example",
            ["module", "submodule", "test"],
            ["another_module"],
            "example"
        ),
        (
            "example",
            ["module", "submodule", "test"],
            ["module.submodule.test"],
            "module.submodule.test.example"
        ),
        (
            "example",
            ["module", "submodule", "test"],
            ["submodule.test"],
            "submodule.test.example"
        ),
        (
            "example",
            ["module", "submodule", "test"],
            ["another_module", "submodule.test", "test"],
            "submodule.test.example"
        )
    ],
    ids=[
        "no module",
        "one module not in hierarchy",
        "one module matching entire hierarchy",
        "one module matching part of the hierarchy",
        "several modules"
    ]
)
def test_guess_module_name(name, hierarchy_folders, module_names, expected):
    """Return module name from initial name, hierarchy folders and modules."""
    assert champollion.parser.js_module._guess_module_name(
        name, hierarchy_folders, module_names
    ) == expected



# :coding: utf-8


def fetch_environment(file_id, files=None, module_names=None):
    """Return module environment dictionary from *file_id*.

    *file_id* represent the identifier of the file.

    *files* is an optional list of the other file names stored in the same
    directory as the one analyzed.

    *module_names* is an optional list of all the other module name
    previously fetched to help determine the module name of the current
    file.

    The environment is in the form of::

        {
            "id": "module.test",
            "name": test,
            "file_id": "module/test/index.js"
        }

    """
    if module_names is None:
        module_names = []

    hierarchy = file_id.split("/")
    file_name = hierarchy.pop()

    if file_name == "index.js":
        module_id = ".".join(hierarchy)
        module_name = _guess_module_name(
            hierarchy[-1],
            hierarchy_folders=hierarchy[:-1],
            module_names=module_names
        )

    elif "index.js" in files:
        name = file_name.split(".js")[0]
        module_id = ".".join(hierarchy + [name])
        module_name = _guess_module_name(
            ".".join([hierarchy[-1], name]),
            hierarchy_folders=hierarchy[:-1],
            module_names=module_names
        )

    else:
        name = file_name.split(".js")[0]
        module_id = ".".join(hierarchy + [name])
        module_name = name

    return {
        "id": module_id,
        "name": module_name,
        "path": module_id.replace(".", "/"),
        "file_id": file_id
    }


def _guess_module_name(name, hierarchy_folders, module_names):
    """Return the full module *name* from *hierarchy_folders*.

    *module_names* is the list of modules already fetched.

    """
    for i in range(len(hierarchy_folders)):
        root_module = ".".join(hierarchy_folders[i:])
        if root_module in module_names:
            return ".".join([root_module, name])

    return name



# :coding: utf-8

import pytest

import champollion.parser


def test_get_environment_error():
    """Raise an error if the path is incorrect."""
    with pytest.raises(OSError):
        champollion.parser.fetch_environment("")


def test_get_environment_empty(temporary_directory):
    """Return an empty environment."""
    environment = {
        "module": {},
        "class": {},
        "method": {},
        "attribute": {},
        "function": {},
        "data": {},
        "file": {}
    }
    assert champollion.parser.fetch_environment(
        temporary_directory
    ) == environment



# :coding: utf-8

from sphinx import addnodes
from docutils import nodes
from sphinx.util.nodes import make_refnode


class ViewCode(object):
    """Helper class to display the code for each :term:`Javascript` module
    found and link it to the documentation.

    """

    #: Store the modules elements to link to the code
    js_modules = None

    @classmethod
    def add_source_code_links(cls, app, doctree):
        """Parse *doctree* and add source code link when available.

        Create temporary 'js_modules' in *app* builder environment to store all
        module information which will be used to create the code page links.

        This function is called with the ``doctree-read`` Sphinx event, emitted
        when a *doctree* has been parsed and read by the  environment, and is
        about to be pickled

        """
        js_env = app.config.js_environment
        builder_env = app.builder.env

        if cls.js_modules is None:
            cls.js_modules = {}

        # Loop through all js signature nodes
        for object_node in doctree.traverse(addnodes.desc):
            if object_node.get("domain") != "js":
                continue

            for node in object_node:
                if not isinstance(node, addnodes.desc_signature):
                    continue

                node_type = node.get("type", None)
                if node_type not in js_env.keys():
                    continue

                node_id = node.get("id", None)
                if node_id not in js_env[node_type].keys():
                    continue

                js_env_element = js_env[node_type][node_id]
                for element in ["name", "module_id", "line_number"]:
                    if element not in js_env_element.keys():
                        continue

                module_id = js_env_element["module_id"]
                page_name = "_modules/{0}".format(module_id.replace(".", "/"))

                if module_id not in cls.js_modules.keys():
                    cls.js_modules[module_id] = {
                        "pagename": page_name,
                        "docname": builder_env.docname,
                        "entries": {}
                    }

                line_number = js_env_element["line_number"]
                cls.js_modules[module_id]["entries"][line_number] = (
                    js_env_element["name"]
                )

                link_node = addnodes.only(expr="html")
                link_node += addnodes.pending_xref(
                    "",
                    reftype="viewcode",
                    refdomain="std",
                    refexplicit=False,
                    reftarget=page_name,
                    refid=node.get("fullname"),
                    refdoc=builder_env.docname
                )
                link_node[0] += nodes.inline(
                    "", "[source]", classes=["viewcode-link"]
                )
                node += link_node

    @classmethod
    def create_code_pages(cls, app):
        """Create all code pages and the links to the documentation.

        This function is called with the ``html-collect-pages`` Sphinx event,
        emitted when the HTML builder is starting to write non-document pages.

        """
        builder_env = app.builder.env
        if cls.js_modules is None:
            return

        module_env = app.config.js_environment["module"]
        file_env = app.config.js_environment["file"]

        highlighter = app.builder.highlighter
        uri = app.builder.get_relative_uri

        all_pages = [elt["pagename"] for elt in cls.js_modules.values()]

        for module_id, element in cls.js_modules.items():
            file_id = module_env[module_id]["file_id"]
            page_name = element["pagename"]
            doc_name = element["docname"]

            if builder_env.config.highlight_language in (
                "js", "default", "none"
            ):
                lexer = builder_env.config.highlight_language
            else:
                lexer = "js"

            highlighted = highlighter.highlight_block(
                file_env[file_id]["content"], lexer, linenos=False
            )
            lines = highlighted.splitlines()

            for line_number, name in element["entries"].items():
                link = uri(page_name, doc_name) + "#" + name
                lines[line_number-1] = (
                    "<div class='viewcode-block' id='{name}'>"
                    "<a class='viewcode-back' href='{link}'>[docs]</a>"
                    "{line}".format(
                        name=name,
                        link=link,
                        line=lines[line_number-1]
                    )
                )

            parents = []

            for module_name in reversed(module_id.split(".")[:-1]):
                link_page = "_modules/{0}".format(module_name.replace(".", "/"))
                if link_page in all_pages:
                    parents.append({
                        "link": uri(page_name, link_page),
                        "title": module_name
                    })

            parents.append({
                "link": uri(page_name, "_modules/index"),
                "title": "Code"
            })
            parents.reverse()

            # putting it all together
            context = {
                "parents": parents,
                "title": module_id,
                "body": (
                    "<h1>Source code for {name}</h1>{content}".format(
                        name=module_id,
                        content="\n".join(lines)
                    )
                ),
            }
            yield page_name, context, "page.html"

        yield cls.create_code_page_index(app)

    @classmethod
    def create_code_page_index(cls, app):
        """Create page index regrouping all code page links.
        """
        if cls.js_modules is None:
            return

        body = ["\n<ul>"]

        for module_id in sorted(cls.js_modules.keys()):
            link_page = "_modules/{0}".format(module_id.replace(".", "/"))
            uri = app.builder.get_relative_uri

            body.append(
                "<li><a href='{link}'>{name}</a></li>\n".format(
                    link=uri("_modules/index", link_page),
                    name=module_id
                )
            )

        body.append("</ul>")

        context = {
            "title": "Overview: module code",
            "body": (
                "<h1>All modules for which code is available</h1>{0}".format(
                    "".join(body)
                )
            ),
        }

        return "_modules/index", context, "page.html"

    @classmethod
    def create_missing_code_link(cls, app, env, node, content_node):
        """Resolve all '[source]' links in Api documentation pages.

        This function is called with the ``missing-reference`` Sphinx event,
        emitted when a cross-reference to a Python module or object cannot be
        resolved.

        """
        if node["reftype"] == "viewcode":
            return make_refnode(
                app.builder,
                node["refdoc"],
                node["reftarget"],
                node["refid"],
                content_node
            )



# :coding: utf-8

import re
import functools

from .helper import filter_comments
from .helper import collapse_all
from .helper import get_docstring


#: Regular Expression pattern for classes
_CLASS_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<export>export +)?(?P<default>default +)?"
    r"(class +(?P<class_name>\w+)|(const|let|var) +(?P<data_name>\w+) "
    r"*= *class +\w+)"
    r"( +extends +(?P<mother_class>[\w._-]+))? *{"
)

#: Regular Expression pattern for class methods
_CLASS_METHOD_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<prefix>(static|get|set) +)?"
    r"(?P<method_name>[\w._-]+) *\([\n ]*(?P<arguments>.*?)[\n ]*\) *{",
    re.DOTALL
)

#: Regular Expression pattern for class arrow methods
_CLASS_METHOD_ARROW_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<prefix>static +)?(?P<method_name>\w+) *= *"
    r"(\([\n ]*(?P<arguments>.*?)[\n ]*\)|(?P<single_argument>[\w._-]+)) *"
    r"=> *",
    re.DOTALL
)

#: Regular Expression pattern for class attribute
_CLASS_ATTRIBUTE_PATTERN = re.compile(
    r"(?P<start_regex>(\n|^)) *(?P<prefix>static +)?"
    r"(?P<name>[\w._-]+) *= *(?P<value>.+?;)",
    re.DOTALL
)


def fetch_environment(content, module_id):
    """Return class environment dictionary from *content*.

    *module_id* represent the identifier of the module.

    The environment is in the form of::

        {
            "moduleName.AwesomeClass": {
                "id": "module.AwesomeClass",
                "name": "AwesomeClass",
                "parent": "MotherClass",
                "line": 42,
                "description": "Class doc.\\n\\nDetailed description."

                "id": "moduleName.AwesomeClass",
                "module_id": "moduleName",
                "exported": False,
                "default": False,
                "name": "AwesomeClass",
                "parent": None,
                "line_number": 2,
                "description": "Class doc.\\n\\nDetailed description."
                "method": {
                    "moduleName.AwesomeClass.awesomeMethod": {
                        ....
                    }
                },
                "attribute": {
                    "moduleName.AwesomeClass.DATA": {
                        ....
                    }
                }
            },
            ...
        }

    """
    environment = {}

    lines = content.split("\n")

    # The comment filter is made during the collapse content process to
    # preserve the class content with all comments (and docstrings!)
    content, collapsed_content = collapse_all(content, filter_comment=True)

    for match in _CLASS_PATTERN.finditer(content):
        class_name = match.group("class_name")
        if class_name is None:
            class_name = match.group("data_name")

        class_id = ".".join([module_id, class_name])

        line_number = (
            content[:match.start()].count("\n") +
            match.group("start_regex").count("\n") + 1
        )

        method_environment = {}
        attribute_environment = {}

        if line_number in collapsed_content.keys():
            class_content = collapsed_content[line_number][1:-1]

            method_environment = fetch_methods_environment(
                class_content, class_id, line_number=line_number-1
            )
            attribute_environment = fetch_attribute_environment(
                class_content, class_id, line_number=line_number-1
            )

        class_environment = {
            "id": class_id,
            "module_id": module_id,
            "exported": match.group("export") is not None,
            "default": match.group("default") is not None,
            "name": class_name,
            "parent": match.group("mother_class"),
            "line_number": line_number,
            "description": get_docstring(line_number, lines),
            "method": method_environment,
            "attribute": attribute_environment
        }
        environment[class_id] = class_environment

    return environment


def fetch_methods_environment(content, class_id, line_number=0):
    """Return function environment dictionary from *content*.

    *class_id* represent the identifier of the method class.

    *line_number* is the first line number of content.

    The environment is in the form of::

        {
            "moduleName.AwesomeClass.awesomeMethod": {
                "id": "moduleName.AwesomeClass.awesomeMethod",
                "class_id": "moduleName.AwesomeClass",
                "module_id": "moduleName",
                "name": "awesomeMethod",
                "prefix": "get",
                "arguments": ["argument1", "argument2"],
                "line_number": 5,
                "description": "Method doc.\\n\\nDetailed description."
            }
        }

    """
    environment = {}

    lines = content.split("\n")
    content = filter_comments(content)
    content = collapse_all(content)[0]

    for match_iter in (
        _CLASS_METHOD_ARROW_PATTERN.finditer(content),
        _CLASS_METHOD_PATTERN.finditer(content)
    ):
        for match in match_iter:
            method_id = ".".join([class_id, match.group("method_name")])
            prefix = match.group("prefix")
            if prefix is not None:
                prefix = prefix.strip()

                # Add the prefix to the method ID if the method if a getter or
                # a setter as several method could have the same name.
                if prefix in ["get", "set"]:
                    method_id += "." + prefix

            _line_number = (
                content[:match.start()].count("\n") +
                match.group("start_regex").count("\n") + 1
            )

            arguments_matched = match.group("arguments")
            if arguments_matched is None:
                arguments_matched = match.group("single_argument")

            arguments = list(filter(lambda x: len(x), [
                arg.strip() for arg in arguments_matched.split(",")
            ]))

            method_environment = {
                "id": method_id,
                "class_id": class_id,
                "module_id": class_id.rsplit(".", 1)[0],
                "name": match.group("method_name"),
                "prefix": prefix,
                "arguments": arguments,
                "line_number": _line_number + line_number,
                "description": get_docstring(_line_number, lines)
            }
            environment[method_id] = method_environment

    return environment


def fetch_attribute_environment(content, class_id, line_number=0):
    """Return attribute environment dictionary from *content*.

    *class_id* represent the identifier of the attribute class.

    *line_number* is the first line number of content.

    The environment is in the form of::

        {
            "moduleName.AwesomeClass.DATA": {
                "id": "moduleName.AwesomeClass.DATA",
                "class_id": "moduleName.AwesomeClass",
                "module_id": "moduleName",
                "name": "DATA",
                "prefix": "static",
                "value": "42",
                "line_number": 8,
                "description": "Attribute doc.\\n\\nDetailed description."
            }
        }

    """
    environment = {}

    lines = content.split("\n")

    # The comment filter is made during the collapse content process to
    # preserve the entire value (with semi-colons and docstrings!)
    content, collapsed_content = collapse_all(content, filter_comment=True)

    for match in _CLASS_ATTRIBUTE_PATTERN.finditer(content):
        attribute_id = ".".join([class_id, match.group("name")])
        prefix = match.group("prefix")
        if prefix is not None:
            prefix = prefix.strip()

        value = match.group("value")

        _line_number = (
            content[:match.start()].count("\n") +
            match.group("start_regex").count("\n") + 1
        )

        for _value_line_number in range(
            _line_number, _line_number + value.count("\n") + 1
        ):
            if "{}" in value and _value_line_number in collapsed_content.keys():
                value = value.replace(
                    "{}", collapsed_content[_value_line_number]
                )

        # Do not keep semi-colon in value
        if value.endswith(";"):
            value = value[:-1]

        attribute_environment = {
            "id": attribute_id,
            "class_id": class_id,
            "module_id": class_id.rsplit(".", 1)[0],
            "name": match.group("name"),
            "prefix": prefix,
            "value": functools.reduce(_clean_value, value.split('\n')).strip(),
            "line_number": _line_number + line_number,
            "description": get_docstring(_line_number, lines)
        }
        environment[attribute_id] = attribute_environment

    return environment


def _clean_value(line1, line2):
    """Clean up attribute value for display."""
    _line1 = line1.strip()
    _line2 = line2.strip()

    # Let trailing space to make the code easier to read
    if _line1[-1:] in ["{", "}", "(", ")", "[", "]", ";", ","]:
        _line1 += " "

    return _line1 + _line2



# :coding: utf-8

import re


#: Regular Expression pattern for single line comments
_ONE_LINE_COMMENT_PATTERN = re.compile(r"(\n|^| )//.*?\n")

#: Regular Expression pattern for multi-line comments
_MULTI_LINES_COMMENT_PATTERN = re.compile(r"/\*.*?\*/", re.DOTALL)

#: Regular Expression pattern for nested element symbols
_NESTED_ELEMENT_PATTERN = re.compile(r"{[^{}]*}")


def filter_comments(
    content, filter_multiline_comment=True, keep_content_size=False
):
    """Return *content* without the comments.

    If *filter_multiline_comment* is set to False, only the one line comment
    will be filtered out.

    If *keep_content_size* is set to True, the size of the content is preserved.

    .. note::

        The filtered content keep the same number of lines as the
        original content.

    .. seealso:: https://www.w3schools.com/js/js_comments.asp

    """
    def _replace_comment(element):
        """Replace matched *element* in content."""
        replacement = ""
        matched = element.group()

        # Ensure that only the comment part is being replaced
        if not matched.startswith("/"):
            replacement += matched[0]
            matched = matched[1:]

        count = matched.count("\n")

        # Add empty spaces with the size of the content if the size
        # must be kept.
        if keep_content_size:
            _buffer = len(matched) - count
            replacement += " " * _buffer + "\n" * count

        # Otherwise simply keep the number of lines
        else:
            replacement += "\n" * count

        return replacement

    content = _ONE_LINE_COMMENT_PATTERN.sub(_replace_comment, content)

    if filter_multiline_comment:
        content = _MULTI_LINES_COMMENT_PATTERN.sub(_replace_comment, content)

    return content


def collapse_all(content, filter_comment=False):
    """Return tuple of *content* with the top level elements only and dictionary
    containing the collapsed content associated with the *line number*.

    If *filter_comment* is set to True, all comment are removed from the content
    before collapsing the elements. The collapsed content dictionary preserve
    the comments.

    .. note::

        The content with collapsed elements keep the same number of
        lines as the original content.

    """
    _initial_content = content
    collapsed_content = {}

    if filter_comment:
        # Filter comment before collapsing elements to prevent comment analysis
        content = filter_comments(content, keep_content_size=True)

    def _replace_element(element):
        """Replace matched *element* in content."""
        # Guess line number
        count = element.group().count("\n")

        # Ensure that the replacement string keep the same length that
        # the original content to be able to use the match positions
        _buffer = len(element.group()) - count - 2

        if len(element.group()) > 2:
            line_number = content[:element.start()].count("\n")+1
            collapsed_content[line_number] = (
                _initial_content[element.start():element.end()]
            )

        return "<>{buffer}{lines}".format(
            buffer=" " * _buffer,
            lines="\n" * count
        )

    _content = None

    while _content != content:
        _content = content
        content = _NESTED_ELEMENT_PATTERN.sub(_replace_element, content)

    # Remove the space buffer before returning the content
    content = re.sub(r"<> *", lambda x: "{}", content)

    return content, collapsed_content


def get_docstring(line_number, lines):
    """Return docstrings for an element at a specific *line_number*.

    Loop into the file *lines* in reverse, starting from the element's
    *line_number* in order to parse the docstring if available.

    The docstring must be in the form of::

        /**
         * Class doc.
         *
         * Detailed description.
         */
        class AwesomeClass {
           ...
        }

    Which will return the following result::

        "Class doc.\\n\\nDetailed description."

    The docstring can also fit on one line, in the form of::

        /** Class doc. */
        class AwesomeClass {
           ...
        }

    """
    docstring = None

    for index in reversed(range(line_number-1)):
        line = lines[index].strip()
        if len(line) == 0 or line.startswith("//"):
            # Do not look for docstring when more than two blank lines precede
            # the element.
            if index < line_number - 1:
                return

            continue

        # Start of the docstring (from the end)
        if docstring is None:
            # If the entire docstring fit in one line
            match = re.search("(?<=/\*\* ).*(?= \*/)", line)
            if match is not None:
                return match.group()

            # No docstring
            if not line.startswith("*/"):
                return

            docstring = []

        # Valid docstring line starts with a '*'
        elif re.search("^\*( *| +.+)$", line) is not None:
            indentation = 2 if len(line) > 1 else 1
            docstring.append(line[indentation:].rstrip())

        # Beginning of valid docstrings starts with '/**'
        elif line.startswith("/**"):
            return "\n".join(docstring[::-1])

        # Error in the docstring
        else:
            return
